@@ -2,13 +2,12 @@
 #![deny(clippy::unwrap_used)]
 #![deny(clippy::panic)]
 
-use anyhow::{anyhow, bail, ensure, Result};
-use cargo_metadata::MetadataCommand;
-use clap::{crate_version, lazy_static::lazy_static, Clap};
+use anyhow::{anyhow, bail, ensure, Context, Result};
 use dylint_internal::{
     env::{self, var},
     Command,
 };
+use lazy_static::lazy_static;
 use std::{
     collections::{BTreeMap, BTreeSet},
     env::consts,
@@ -20,6 +19,12 @@ use std::{
 
 pub mod driver_builder;
 
+#[cfg(feature = "metadata")]
+mod metadata;
+
+#[cfg(feature = "metadata")]
+mod toml;
+
 mod error;
 use error::warn;
 pub use error::{ColorizedError, ColorizedResult};
@@ -36,99 +41,43 @@ type ToolchainMap = BTreeMap<String, BTreeSet<PathBuf>>;
 
 pub type NameToolchainMap = BTreeMap<String, ToolchainMap>;
 
-#[derive(Clap, Debug)]
-#[clap(bin_name = "cargo")]
-struct Opts {
-    #[clap(subcommand)]
-    subcmd: SubCommand,
-}
-
-#[derive(Clap, Debug)]
-enum SubCommand {
-    Dylint(Dylint),
-}
-
-#[derive(Clap, Debug, Default)]
-#[clap(
-    version = crate_version!(),
-    after_help = r#"ENVIRONMENT VARIABLES:
-
-DYLINT_DRIVER_PATH (default: $HOME/.dylint_drivers) is the directory where Dylint stores rustc
-drivers.
-
-DYLINT_LIBRARY_PATH (default: none) is a colon-separated list of directories where Dylint searches
-for libraries.
-
-DYLINT_RUSTFLAGS (default: none) is a space-separated list of flags that Dylint passes to `rustc`.
-"#,
-)]
+#[allow(clippy::struct_excessive_bools)]
+#[derive(Debug, Default)]
 pub struct Dylint {
-    #[clap(long = "all", about = "Load all discovered libraries")]
     pub all: bool,
-
-    #[clap(
-        multiple = true,
-        number_of_values = 1,
-        long = "lib",
-        value_name = "name",
-        about = "Library name to load lints from. A file with a name of the form \"DLL_PREFIX \
-        <name> '@' TOOLCHAIN DLL_SUFFIX\" is searched for in the directories listed in \
-        DYLINT_LIBRARY_PATH and in the current package's `debug` and `release` directories."
-    )]
     pub libs: Vec<String>,
-
-    #[clap(
-        long = "list",
-        about = "If no libaries are named, list the name, toolchain, and location of all \
-        discovered libraries. If at least one library is named, list the name, level, and \
-        description of all lints in all named libraries. Combine with `--all` to list all \
-        lints in all discovered libraries."
-    )]
     pub list: bool,
-
-    #[clap(
-        multiple = true,
-        number_of_values = 1,
-        long = "path",
-        value_name = "path",
-        about = "Library path to load lints from"
-    )]
+    pub manifest_path: Option<String>,
+    pub no_build: bool,
+    pub no_metadata: bool,
+    pub packages: Vec<String>,
     pub paths: Vec<String>,
-
-    #[clap(short = 'q', long = "quiet", about = "Suppress warnings")]
     pub quiet: bool,
-
-    #[clap(
-        about = "Libraries to load lints from. Each <name> is searched for as described under \
-        `--lib`. If no library is found, <name> is treated as path. To avoid ambiguity, use \
-        `--lib` or `--path`."
-    )]
+    pub workspace: bool,
     pub names: Vec<String>,
-
-    #[clap(last = true, about = "Arguments for `cargo check`")]
     pub args: Vec<String>,
 }
 
-pub fn cargo_dylint<T: AsRef<OsStr>>(args: &[T]) -> ColorizedResult<()> {
-    match Opts::parse_from(args).subcmd {
-        SubCommand::Dylint(opts) => run(&opts),
-    }
-    .map_err(ColorizedError::new)
-}
-
 pub fn run(opts: &Dylint) -> Result<()> {
-    let name_toolchain_map = name_toolchain_map()?;
+    let name_toolchain_map = name_toolchain_map(opts)?;
 
     run_with_name_toolchain_map(opts, &name_toolchain_map)
 }
 
-pub fn name_toolchain_map() -> Result<NameToolchainMap> {
+#[allow(unused_variables)]
+pub fn name_toolchain_map(opts: &Dylint) -> Result<NameToolchainMap> {
     let mut name_toolchain_map = NameToolchainMap::new();
 
     let dylint_library_paths = dylint_library_paths()?;
-    let profile_paths = profile_paths();
+    #[cfg(feature = "metadata")]
+    let workspace_metadata_paths = metadata::workspace_metadata_paths(opts)?;
+    #[cfg(not(feature = "metadata"))]
+    let workspace_metadata_paths = vec![];
 
-    for path in dylint_library_paths.iter().chain(&profile_paths) {
+    for (path, require_existence) in dylint_library_paths.iter().chain(&workspace_metadata_paths) {
+        if !require_existence && !path.exists() {
+            continue;
+        }
         for entry in dylint_libraries_in(path)? {
             let (name, toolchain, path) = entry?;
             name_toolchain_map
@@ -143,7 +92,7 @@ pub fn name_toolchain_map() -> Result<NameToolchainMap> {
     Ok(name_toolchain_map)
 }
 
-fn dylint_library_paths() -> Result<Vec<PathBuf>> {
+fn dylint_library_paths() -> Result<Vec<(PathBuf, bool)>> {
     let mut paths = Vec::new();
 
     if let Ok(val) = var(env::DYLINT_LIBRARY_PATH) {
@@ -159,35 +108,19 @@ fn dylint_library_paths() -> Result<Vec<PathBuf>> {
                 "DYLINT_LIBRARY_PATH contains `{}`, which is not a directory",
                 path.to_string_lossy()
             );
-            paths.push(path);
+            paths.push((path, true));
         }
     }
 
     Ok(paths)
 }
 
-fn profile_paths() -> Vec<PathBuf> {
-    let mut paths = Vec::new();
-
-    if let Ok(metadata) = MetadataCommand::new().no_deps().exec() {
-        let debug = metadata.target_directory.join_os("debug");
-        let release = metadata.target_directory.join_os("release");
-        if debug.is_dir() {
-            paths.push(debug);
-        }
-        if release.is_dir() {
-            paths.push(release);
-        }
-    }
-
-    paths
-}
-
 #[allow(clippy::option_if_let_else)]
 fn dylint_libraries_in(
     path: &Path,
 ) -> Result<impl Iterator<Item = Result<(String, String, PathBuf)>>> {
-    let iter = read_dir(path)?;
+    let iter = read_dir(path)
+        .with_context(|| format!("`read_dir` failed for `{}`", path.to_string_lossy()))?;
     Ok(iter
         .map(|entry| -> Result<Option<(String, String, PathBuf)>> {
             let entry = entry?;
@@ -418,7 +351,7 @@ fn parse_target_name(target_name: &str) -> Option<(String, String)> {
 }
 
 fn list_lints(
-    _opts: &Dylint,
+    opts: &Dylint,
     name_toolchain_map: &NameToolchainMap,
     resolved: &ToolchainMap,
 ) -> Result<()> {
@@ -429,7 +362,7 @@ fn list_lints(
                     .get(toolchain)
                     .map_or(false, |paths| paths.contains(path))
                 {
-                    let driver = driver_builder::get(toolchain)?;
+                    let driver = driver_builder::get(opts, toolchain)?;
                     let dylint_libs = serde_json::to_string(&[path])?;
 
                     print!("{}", name);
@@ -471,8 +404,19 @@ fn check(
     let clippy_disable_docs_links = clippy_disable_docs_links()?;
 
     for (toolchain, paths) in resolved {
-        let driver = driver_builder::get(toolchain)?;
+        let driver = driver_builder::get(opts, toolchain)?;
         let dylint_libs = serde_json::to_string(&paths)?;
+        let mut args = vec![];
+        if let Some(path) = &opts.manifest_path {
+            args.extend(&["--manifest-path", path]);
+        }
+        for spec in &opts.packages {
+            args.extend(&["-p", spec]);
+        }
+        if opts.workspace {
+            args.extend(&["--workspace"]);
+        }
+        args.extend(opts.args.iter().map(String::as_str));
 
         // smoelius: Set CLIPPY_DISABLE_DOCS_LINKS to prevent lints from accidentally linking to the
         // Clippy repository. But set it to the JSON-encoded original value so that the Clippy
@@ -495,7 +439,7 @@ fn check(
                 ),
                 (env::RUSTUP_TOOLCHAIN.to_owned(), toolchain.clone()),
             ])
-            .args(opts.args.clone())
+            .args(args)
             .success()?;
     }
 
@@ -524,7 +468,7 @@ mod test {
     #![allow(clippy::unwrap_used)]
 
     use super::*;
-    use dylint_examples as examples;
+    use dylint_internal::examples;
     use lazy_static::lazy_static;
     use std::env::set_var;
     use test_env_log::test;
@@ -545,7 +489,11 @@ mod test {
                 .collect::<Vec<_>>()
                 .join(":");
             set_var(env::DYLINT_LIBRARY_PATH, dylint_library_path);
-            name_toolchain_map().unwrap()
+            name_toolchain_map(&Dylint {
+                no_metadata: true,
+                ..Dylint::default()
+            })
+            .unwrap()
         };
     }
 
@@ -572,19 +520,19 @@ mod test {
         run_with_name_toolchain_map(&opts, &NAME_TOOLCHAIN_MAP).unwrap();
     }
 
-    // smoelius: Check that loading multiple libraries with the same Rust toolchain works. At one point,
-    // I was getting this error from `libloading`:
+    // smoelius: Check that loading multiple libraries with the same Rust toolchain works. At one
+    // point, I was getting this error from `libloading`:
     //
     //   cannot allocate memory in static TLS block
     //
-    // The culprit turned out to be the `rand` crate, which uses a lot of thread local storage. `rand`
-    // is used by `tempfile`, which is used by various Rust compiler crates. Essentially, each library
-    // had its own copy of the Rust compiler, and therefore its own copy of the `rand` crate, and this
-    // was eating up all the thread local storage.
+    // The culprit turned out to be the `rand` crate, which uses a lot of thread local storage.
+    // `rand` is used by `tempfile`, which is used by various Rust compiler crates. Essentially,
+    // each library had its own copy of the Rust compiler, and therefore its own copy of the `rand`
+    // crate, and this was eating up all the thread local storage.
     //
-    // The solution was to add `extern crate rustc_driver` to each library. This causes the library to
-    // link against `librust_driver.so`, which dylint-driver also links against. So, essentially, each
-    // library now uses dylint-driver's copy of the `rand` crate.
+    // The solution was to add `extern crate rustc_driver` to each library. This causes the library
+    // to link against `librust_driver.so`, which dylint-driver also links against. So, essentially,
+    // each library now uses dylint-driver's copy of the `rand` crate.
     //
     // This thread was very helpful in diagnosing the problem:
     //
